# -*- coding: utf-8 -*-
"""
This module contains functions for working with fabric objects
of the HPE Composabale Fabric Manager instance.
"""

####################
# Fabric functions #
####################

def get_fabrics(cfmclient, fabric_uuid=None):
    """
    Get a list of Fabrics currently defined in Composable Fabric.

    :param cfmclient:
    :param fabric_uuid: switch_uuid: UUID of switch from which to fetch port data
    :return: list of Dictionary objects where each dictionary represents a port on a
    Composable Fabric Module
    :rtype: list
    """
    path = 'v1/fabrics'
    if fabric_uuid:
        path += '/{}'.format(fabric_uuid)

    return cfmclient.get(path).json().get('result')


def create_fabric(cfmclient, data, fabric_type=None):
    """
    Create :A Composable Fabric.
    """
    path = 'fabrics'
    params = {'type': fabric_type} if fabric_type else None

    return cfmclient.post(path, params, data).json().get('result')


def get_fabric_ip_networks(cfmclient, fabric_uuid=None):
    """
    Get a list of IP networks from the Composable Fabric.

    :param cfmclient:
    :param fabric_uuid: UUID of fabric
    :return: list of IP address dict objects
    :rtype: list
    """
    path = 'v1/fabric_ip_networks'
    if fabric_uuid:
        path += '/{}'.format(fabric_uuid)

    return cfmclient.get(path).json().get('result')


#####################
# Fitting functions #
#####################

def perform_fit(cfmclient, fabric_uuid, name, description):
    """
    Request a full fit across managed Composable Fabrics.

    :param cfmclient:
    :return:
    """
    data = {
        'fabric_uuid': '{}'.format(fabric_uuid),
        'name': '{}'.format(name),
        'description': '{}'.format(description)
    }
    path = 'fits'

    return cfmclient.post(path, data)


####################
# Switch functions #
####################

def get_switches(cfmclient, params=None):
    """
    Get a list of Composable Fabric switches

    :param cfmclient: object of type CFMClient
    :param params: dict of query parameters used to filter request from API
    :return: list of dicts

    >>> client= CFMClient('10.101.0.210', 'admin', 'plexxi')
    >>> get_switches(client, params={'ports': True})
    >>> get_switches(client, params={'ports': True, 'software': True})
    >>> get_switches(client, params={'ports': True, 'software': True, 'fabric': True})
    """

    return cfmclient.get('v1/switches', params).json().get('result')


def create_switch(cfmclient, data, fabric_type=None):
    """
    Create a Composable Fabric switch.
    """
    params = {'type': fabric_type} if fabric_type else None

    return cfmclient.post('v1/switches', params, data).json().get('result')


##################
# Port functions #
##################

def get_ports(cfmclient, switch_uuid=None):
    """
    Get Composable Fabric switch ports.

    :param cfmclient: object of type CFMClient
    :param switch_uuid: switch_uuid: UUID of switch from which to fetch port data
    :return: list of Dictionary objects where each dictionary represents a port on a
    Composable Fabric Module
    :rtype: list
    """
    path='v1/ports'
    if switch_uuid:
        path += '?switches={}&type=access'.format(switch_uuid)

    return cfmclient.get(path).json().get('result')


def update_ports(cfmclient, port_uuids, field, value):
    """
    Update attributes of composable fabric switch ports

    :param cfmclient:
    :param port_uuids: list of str representing Composable Fabric port UUIDs
    :param field: str specific field which is desired to be modified (case-sensitive)
    :param value: str specific field which sets the new desired value for the field
    :return: dict which contains count, result, and time of the update
    :rtype: dict
    """
    if port_uuids:
        data = [{
            'uuids': port_uuids,
            'patch': [
                {
                    'path': '/{}'.format(field),
                    'value': value,
                    'op': 'replace'
                }
            ]
        }]
        cfmclient.patch('v1/ports', data)

##################
# VLAN functions #
##################

def get_vlan_groups(cfmclient, params=None):
    """
    Get Composable Fabric vlan groups.

    :param cfmclient: object of type CFMClient
    :return: list of VLAN Group dictionary objects in the Composable Fabric
    :rtype: list
    """

    return cfmclient.get('v1/vlan_groups', params).json().get('result')


# TODO POST VLAN_GROUP FUNCTION

# TODO PUT VLAN GROUP FUNCTION

# TODO DELETE VLAN GROUP FUNCTION


def get_vlan_properties(cfmclient, fabric_uuid):
    """
    Get Composable Fabric vlan properties for a specific fabric.

    :param cfmclient: object of type CFMClient
    :param fabric_uuid: str representing a valid fabric UUID
    :return: list of VLAN Group Property dictionary objects from the Composable Fabric
    :rtype: list
    """
    path='v1/vlan_properties/{}'.format(fabric_uuid)

    return cfmclient.get(path).json().get('result')

####################
# VPC functions #
####################

def get_vpcs(cfmclient, uuid=None):
    """
    Get a list of VPCs currently defined in Composable Fabric.

    :param cfmclient: Connected CFM API client
    :param uuid: specific VPC UUID to retrieve
    :return: list of VPC dictionary objects
    :rtype: list
    """
    path = 'v1/vpcs'
    if uuid:
        path += '/{}'.format(uuid)

    return cfmclient.get(path).json().get('result')

def get_bgp(cfmclient, uuid):
    """
    Get VPC BGP configuration.

    :param cfmclient: Connected CFM API client
    :param uuid: VPC UUID from which to retrieve BGP info
    :return: BGP dictionary object
    :rtype: list
    """
    path = 'v1/vpcs/{}/bgp'.format(uuid)

    return cfmclient.get(path).json().get('result')

def get_bgp_leaf_spine(cfmclient, uuid):
    """
    Get VPC BGP leaf and spine configuration.

    :param cfmclient: Connected CFM API client
    :param uuid: VPC UUID from which to retrieve BGP info
    :return: BGP dictionary object
    :rtype: list
    """
    path = 'v1/vpcs/{}/bgp/leaf_spine'.format(uuid)

    return cfmclient.get(path).json().get('result')

def update_bgp_leaf_spine(cfmclient, uuid, config):
    """
    Update VPC BGP leaf and spine configuration.

    :param cfmclient: Connected CFM API client
    :param uuid: VPC UUID from which to retrieve BGP info
    :param config: BGP leaf spine configuration data
    :return: BGP dictionary object
    :rtype: list
    """
    path = 'v1/vpcs/{}/bgp/leaf_spine'.format(uuid)

    return cfmclient.put(path, config).json().get('result')



# -*- coding: utf-8 -*-
"""
This module provides helper functions and holds the main client object
for authenticating with an HPE Composable Fabric Manager.
"""
import time

import requests

# The following lines remove warnings for self-signed certificates
# noinspection PyUnresolvedReferences
from requests.packages.urllib3.exceptions import \
    InsecureRequestWarning  # pylint: disable=import-error
# noinspection PyUnresolvedReferences
requests.packages.urllib3.disable_warnings(InsecureRequestWarning)  # pylint: disable=no-member


class CFMApiError(Exception):
    """Composable Fabric Manager API exception."""
    pass


class CFMClient(object):
    """Client class for the CFM REST API bindings."""

    def __init__(self, host, username, password, verify_ssl=False, timeout=30):
        """
        Initialize API instance.

        :param host: str FQDN or IPv4 Address of the target CFM host
        :param username: str valid username with sufficient permissions on the CFM host
        :param password: str valid password for username var
        :param verify_ssl: bool verify SSL certificate. Default value is False
        :param timeout: int timeout in seconds for API calls
        """
        self._host = host
        self._username = username
        self._password = password
        self._verify_ssl = verify_ssl
        self._timeout = timeout
        self._session = None
        self._auth_token = None
        self._max_connection_retries = 3

    def __del__(self):
        """
        Disconnect from API on instance destruction.
        Explicitly expires token on CFM server instance.
        """
        self.disconnect()

    def connect(self, login=True):
        """
        Connect to CFM API and retrieve token.

        :param login: bool login and retrieve authentication token
        """
        self._session = None
        self._auth_token = None

        self._session = requests.session()
        self._session.headers.update({'Accept': 'application/json'})
        self._session.headers.update({'Content-Type': 'application/json'})

        if login:
            self._session.headers.update({'X-Auth-Username': '{}'.format(self._username)})
            self._session.headers.update({'X-Auth-Password': '{}'.format(self._password)})

            response = self._call_api('POST', 'v1/auth/token').json()
            self._auth_token = response.get('result')
            if self._auth_token:
                self._session = requests.session()
                self._session.headers.update({'Accept': 'application/json'})
                self._session.headers.update({'Authorization': 'Bearer {}'.format(self._auth_token)})
                self._session.headers.update({'X-Auth-Refresh-Token': 'true'})
            else:
                raise CFMApiError('Error retrieving authentication token')

    def disconnect(self):
        """Disconnect from CFM API session and delete token."""
        self._auth_token = None
        self._session = None

    def delete(self, path, params= None):
        """
        Helper function for HTTP DELETE commands

        :param path: str the requested path
        :return: requests.Response API call response
        :rtype: requests.Response
        """
        return self._call_api(method='DELETE', path=path, params= params)

    def get(self, path, params= None):
        """
        Helper function for HTTP GET commands

        :param path: str the requested path
        :return: requests.Response API call response
        :rtype: requests.Response
        """
        return self._call_api(method='GET', path=path, params=params)

    def patch(self, path, data):
        """
        Helper function for HTTP PATCH commands

        :param path: str the requested path
        :param data: dict the data to send
        :return: requests.Response API call response
        :rtype: requests.Response
        """
        return self._call_api(method='PATCH', path=path, json=data)

    def post(self, path, params=None, data=None):
        """
        Helper function for HTTP POST commands

        :param path: str the requested path
        :param json: dict the data to send
        :return: requests.Response API call response
        :rtype: requests.Response
        """
        return self._call_api(method='POST', path=path, params=params, json=data)

    def put(self, path, data):
        """
        Helper function for HTTP PUT commands

        :param path: str the requested path
        :param data: dict the data to send
        :return: requests.Response API call response
        :rtype: requests.Response
        """
        return self._call_api(method='PUT', path=path, json=data)

    def _call_api(self, method, path, params=None, headers=None, json=None,
                  timeout=None, stream=False):
        """Execute a CFM REST API request.

        Arguments:
            method (str): HTTP request type
            path (str): The request path of the REST API.
            params (dict): (Optional) query parameters.
            headers (dict): (Optional) HTTP Headers to send with the request.
            json (dict): (Optional) json to send in the body of the request.
            timeout (int): Optional timeout override.

        Returns:
            requests.Response (class): JSON representation of the response object from requests
        """
        request_headers = headers if headers else {'Content-Type': 'application/json;charset=UTF-8'}

        if self._auth_token:
            # Set Auth Token in Header
            request_headers.update(
                {'Authorization': 'Bearer {}'.format(self._auth_token),
                 'X-Auth-Refresh-Token': 'true'}
            )
        elif path != 'v1/auth/token':
            # If this is not a login request, then there is a problem with the auth token.
            print('{} {} Aborted: Failed to obtain auth token.'.format(method, path))
            return

        attempts = 0
        while attempts < self._max_connection_retries:
            try:
                attempts += 1
                if self._session:
                    return self._process_request(
                        self._session, method, path, params, request_headers, json, timeout,
                        verify=self._verify_ssl, stream=stream)
                else:
                    with requests.session() as session:
                        session.headers.update({'Accept': 'application/json'})
                        return self._process_request(
                            session, method, path, params, request_headers, json, timeout,
                            verify=self._verify_ssl, stream=stream)
            except requests.exceptions.ConnectionError as exception:
                print('Request failed with error %s', exception)
                if attempts >= self._max_connection_retries:
                    raise exception
                else:
                    print('Retrying Connect API request')
            except requests.exceptions.HTTPError as exception:
                error_code = int(exception.response.status_code)
                if error_code == 401:
                    print('API token no longer valid')
                    self.connect()
                    request_headers.update({'Authorization': 'Bearer {}'.format(self._auth_token)})
                    print('Retrying request %s:%s', method, path)
                elif error_code == 503:
                    print('Service Unavailable on  %s :%s ', method, path)
                    if attempts < self._max_connection_retries:
                        time.sleep(10)
                    else:
                        print('Raise exception retried %s times', attempts)
                        raise exception
                else:
                    print('Exception in API call: %s', exception)
                    raise exception
            except requests.exceptions.ReadTimeout as exception:
                print('ReadTimeout in API call attempt ( %s ) : %s', attempts,
                                   exception)
                if method in ['GET', 'get'] and attempts < self._max_connection_retries:
                    #  Sleep for 2 seconds before re-trying
                    #  Using native sleep method as opposed to eventlet.sleep as we are
                    #  dealing with requests module.
                    time.sleep(2)
                else:
                    print('Retry count ( %s )', attempts)
                    raise exception
                print('Retrying request %s : %s', method, path)
            except Exception as exception:
                print('Exception in API call: %s', exception)
                raise exception

    def _process_request(self, session, method, path, params, headers, json, timeout=None,
                         verify=False, cert=None, stream=False):
        """Execute a REST API request using the supplied session.

        Arguments:
            session (requests.Session): The session to use to issue the HTTP request.
            method (str): HTTP request type
            path (str): The request path of the Prism REST API.
            params (dict): Optional query parameters.
            headers (dict): Optional HTTP Headers to send with the request.
            json (dict): Optional json to send in the body of the request.
            timeout (int): Optional timeout override.
            verify (str): Optional path to trusted CA and/or server certificate(s).
            cert (tuple): Optional client private key and certificate (for two-way TLS).

        Raises (Exception):
            For any exception encountered.

        Returns:
            Response: The response object from requests
        """
        url = 'https://{}/api/{}'.format(self._host, path)
        request = {
            'method': method,
            'url': url,
            'timeout': timeout or self._timeout,
            'headers': headers,
            'params': params,
            'json': json,
            'verify': verify,
            'cert': cert,
            'stream': stream,
        }

        try:
            response = session.request(**request)
        except requests.exceptions.SSLError as exception:
            print('%s %s failed due to SSL handshake error; ensure %s contains a '
                               'certificate which can be used to validate HTTPS connections to %s',
                               method, path, verify, self._host)
            raise exception

        try:
            response.raise_for_status()
            return response
        except Exception as exception:
            print('%s %s failed with status %s',
                  method, path, response.status_code)
            raise exception



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This module contains functions related to working with the system characteristics of the
desired HPE Composable Fabric Manager instance
"""


def get_versions(cfmclient):
    """
    Function takes input cfmclient type object to authenticate against CFM API and queries
    versions API to return the version number of the system represented by the CFCMclient object
    Current supported params are

    :param cfmclient: object of type CFMClient
    :return: list of dicts
    """
    path = 'versions'
    response = cfmclient.get(path)
    return response.json().get('result') if response else None


def get_audit_logs(cfmclient):
    """
    Get :List of audit log records currently defined in Composable Fabric.
    :param cfmclient:
    :return: list of Dictionary objects where each dictionary represents a single entry in the
    audit log of the HPE Composable Fabric Manager
    :rtype: list
    """
    path = 'v1/audits'
    response = cfmclient.get(path)
    return response.json().get('result') if response else None


def get_backups(cfmclient, uuid=None):
    """
    Function to get a list of current backups located on the Composable Fabric Manager
    represented by the cfcmclient object.
    :param cfcmclient: Composable Fabric Manager connection object of type CFMClient
    :param uuid: str that represents the unique identifier for a specific backup
    :return: single dictionary if UUID. List of dictionaries where each dictionary represents a
    single backup.
    """
    path = 'v1/backups'
    if uuid:
        path += '/{}'.format(uuid)
    response = cfmclient.get(path)
    return response.json().get('result') if response else None


def create_backup(cfmclient):
    """
    Function to initiate a new backup on the Composable Fabric Manager represented by the
    CFCMClinet object
    :param cfcmclient:Composable Fabric Manager connection object of type CFMClient
    :return: HTTP response
    """
    path = 'v1/backups'
    response = cfmclient.get(path)
    return response.json().get('result') if response else None
